@@ -309,12 +309,18 @@ resolve_async (YelpUri *uri)
     }
     else if (strchr (priv->res_arg, ':')) {
         priv->tmptype = YELP_URI_DOCUMENT_TYPE_EXTERNAL;
-        priv->fulluri = g_strdup (priv->res_arg);
     }
     else {
         resolve_file_path (uri);
     }
 
+    /* We _always_ want to have a non-null fulluri, so check for it
+     * having been set here and, if we can't think of something
+     * better, set it to res_arg. */
+    if (!priv->fulluri) {
+        priv->fulluri = g_strdup (priv->res_arg);
+    }
+
  done:
     g_idle_add ((GSourceFunc) resolve_final, uri);
 }
@@ -368,6 +374,15 @@ yelp_uri_get_document_uri (YelpUri *uri)
     YelpUriPrivate *priv = GET_PRIV (uri);
     if (priv->doctype == YELP_URI_DOCUMENT_TYPE_UNRESOLVED)
         return NULL;
+
+    /* There's some client code where it makes sense to want a
+     * document uri, whether or not it conforms to a scheme we really
+     * understand. For example, we might want to look up whether the
+     * given page is currently being visited. */
+    if ((!priv->docuri) && priv->fulluri) {
+        return g_strdup (priv->fulluri);
+    }
+
     return g_strdup (priv->docuri);
 }
 
@@ -579,12 +594,22 @@ resolve_data_dirs (YelpUri      *ret,
             }
             g_free (filename);
 
-            filename = g_strdup_printf ("%s/%s.xml", helpdir, pageid);
-            if (g_file_test (filename, G_FILE_TEST_IS_REGULAR)) {
-                priv->tmptype = YELP_URI_DOCUMENT_TYPE_DOCBOOK;
-                continue;
+            if (langfirst) {
+                filename = g_strdup_printf ("%s/index.docbook", helpdir);
+                if (g_file_test (filename, G_FILE_TEST_IS_REGULAR)) {
+                    priv->tmptype = YELP_URI_DOCUMENT_TYPE_DOCBOOK;
+                    continue;
+                }
+                g_free (filename);
+            }
+            else {
+                filename = g_strdup_printf ("%s/%s.xml", helpdir, pageid);
+                if (g_file_test (filename, G_FILE_TEST_IS_REGULAR)) {
+                    priv->tmptype = YELP_URI_DOCUMENT_TYPE_DOCBOOK;
+                    continue;
+                }
+                g_free (filename);
             }
-            g_free (filename);
 
             filename = g_strdup_printf ("%s/%s.html", helpdir, pageid);
             if (g_file_test (filename, G_FILE_TEST_IS_REGULAR)) {
@@ -1271,7 +1296,9 @@ resolve_gfile (YelpUri *uri, const gchar *hash)
     if (priv->fulluri == NULL)
         priv->fulluri = g_strconcat (priv->docuri,
                                      (priv->page_id || priv->frag_id) ? "#" : NULL,
-                                     priv->page_id ? priv->page_id : priv->frag_id,
+                                     priv->page_id ? priv->page_id : "",
+                                     priv->frag_id ? "#" : NULL,
+                                     priv->frag_id ? priv->frag_id : NULL,
                                      NULL);
 
     g_object_unref (info);
